/** @file wlan_mac_queue.h
 *  @brief Transmit Queue Framework
 *
 *  This contains code for accessing the transmit queue.
 *
 *  @copyright Copyright 2013-2018, Mango Communications. All rights reserved.
 *          Distributed under the Mango Communications Reference Design License
 *              See LICENSE.txt included in the design archive or
 *              at http://mangocomm.com/802.11/license
 *
 *  This file is part of the Mango 802.11 Reference Design (https://mangocomm.com/802.11)
 */

/*************************** Constant Definitions ****************************/
#ifndef WLAN_MAC_QUEUE_H_
#define WLAN_MAC_QUEUE_H_

/***************************** Include Files *********************************/

#include "wlan_mac_high_sw_config.h"
#include "xil_types.h"
#include "wlan_common_types.h"
#include "wlan_mac_802_11_defs.h"

//Forward declarations
struct station_info_t;

//-----------------------------------------------
// Queue defines
//
#define QUEUE_ELEMENT_SIZE 0x1000    // 4KB

typedef struct pyld_queue_hdr_t{
        u32	buffer_length; // Number of bytes until next pyld_queue_hdr_t. Currently fixed at QUEUE_ELEMENT_SIZE
        //---- 4-bytes ----
        dl_entry* dle;
        //---- 8-bytes ----
} pyld_queue_hdr_t;

ASSERT_TYPE_SIZE(pyld_queue_hdr_t, 8);

// Base stencil for every Pkt Queue Buffer
typedef struct pkt_queue_buffer_t{
	pyld_queue_hdr_t pyld_queue_hdr;
	u8 reserved[QUEUE_ELEMENT_SIZE - sizeof(pyld_queue_hdr_t)];
} pkt_queue_buffer_t;

ASSERT_TYPE_SIZE(pkt_queue_buffer_t, QUEUE_ELEMENT_SIZE);

// Stencil to use Pkt Queue Buffer for Tx Queue Buffer
typedef struct tx_queue_buffer_t{
	pyld_queue_hdr_t pyld_queue_hdr;
	//---- 8-bytes ----
	u16 length;
    u16 c0_len;
    //---- 12-bytes ----
    u16 c1_len;
    u8 c1_offset;
    u8 reserved0;
    //---- 16-bytes ----
	struct station_info_t* station_info;
	//---- 20-bytes ----
	tx_queue_details_t tx_queue_details;
	//---- 32-bytes ----
    u16 flags;
    u8 reserved1[6];
    //---- 40-bytes ----
	u8 pkt[QUEUE_ELEMENT_SIZE - 40];
} tx_queue_buffer_t;

ASSERT_TYPE_SIZE(tx_queue_buffer_t, QUEUE_ELEMENT_SIZE);

// Stencil to use Pkt Queue Buffer as Rx Eth Queue Buffer
typedef struct rx_eth_queue_buffer_t{
	pyld_queue_hdr_t pyld_queue_hdr;
	//---- 8-bytes ----
	u16 length;
	u8 reserved[ 30 ];
	//---- 40-bytes ----
	u8 encap_eth_header_placeholder[ sizeof(mac_header_80211) + sizeof(llc_header_t) ];
	//---- 72-bytes ----
	u8 pkt[QUEUE_ELEMENT_SIZE - 72];
} rx_eth_queue_buffer_t;

ASSERT_TYPE_SIZE(rx_eth_queue_buffer_t, QUEUE_ELEMENT_SIZE);
ASSERT_MEMBER_ALIGNMENT(rx_eth_queue_buffer_t, pkt, 8); //ensure 8-byte alignment on Ethernet frame for DMA

#define TX_QUEUE_BUFFER_FLAGS_FILL_TIMESTAMP	0x0001
#define TX_QUEUE_BUFFER_FLAGS_FILL_DURATION	0x0002
#define TX_QUEUE_BUFFER_FLAGS_FILL_UNIQ_SEQ	0x0004
#define TX_QUEUE_BUFFER_FLAGS_BYPASS_RECOVERY	0x0008
#define TX_QUEUE_BUFFER_FLAGS_TYPE_CTS 		0x0010
#define TX_QUEUE_BUFFER_FLAGS_TYPE_ACK 		0x0020


/*************************** Function Prototypes *****************************/

void                queue_init();

void                enqueue_after_tail(u16 queue_sel, dl_entry* tqe);
dl_entry* 			dequeue_from_head(u16 queue_sel);
void 				transmit_checkin(dl_entry* tx_queue_buffer_entry);
void 	    		queue_set_state_change_callback(function_ptr_t callback);

dl_entry* 			queue_checkout();
void                queue_checkin(dl_entry* tqe);

int                 queue_checkout_list(dl_list* new_list, u16 num_tqe);
int                 queue_checkin_list(dl_list * list);

u32          		queue_num_free();
u32          		queue_num_queued(u16 queue_sel);
u32                 queue_total_size();

void                purge_queue(u16 queue_sel);

#endif /* WLAN_MAC_QUEUE_H_ */
