function pkt = util_dsss_tx(samp_rate, mac_payload, payload_rate)
% dsss_tx(samp_rate, mac_payload) Computes 802.11b 1Mbps DSSS waveform for
%  given sample rate and MAC payload
%
% samp_rate: Sample rate in MSps; must be 20 or 40
%
% mac_payload: 1-D array of 8-bit integers, used as MAC payload
%  The length of the mac_payload array will be used as the LENGTH value
%  for the comptued waveform. The last 4 bytes of the mac_payload array 
%  will be overwritten with the 4-byte FCS computed over the payload.

pkt = struct();

SCR_INIT_STATE_11B_SPEC = bin2dec('1101100');
SCR_INIT_STATE  = SCR_INIT_STATE_11B_SPEC;

tx_sim = struct();
tx_sim.mac_payload = [mac_payload(:).', zeros(1, 4)];
tx_sim.mac_payload_length = length(tx_sim.mac_payload); %includes FCS

PHY_HDR_SYNC = ones(1, 128);
PHY_HDR_SFD = fliplr(util_hex_to_bin('F3A0', 16));
PHY_HDR_SERVICE = [0 0 1 0 0 0 0 0]; %11b sets b2 for LOCKED_CLOCKS

if(payload_rate == 1)
    PHY_HDR_SIGNAL = fliplr([0 0 0 0 1 0 1 0]); %0x0A for 1Mbps
    phy_hdr_length = fliplr(util_dec_to_bin_vec(8*tx_sim.mac_payload_length, 16));
elseif(payload_rate == 2)
    PHY_HDR_SIGNAL = fliplr([0 0 0 1 0 1 0 0]); %0x14 for 2Mbps
    phy_hdr_length = fliplr(util_dec_to_bin_vec(4*tx_sim.mac_payload_length, 16));
else
    error('Invalid payload_rate value %d - must be 1 or 2', payload_rate)
end

% Calculate the PHY header checksum
phy_hdr_protected = [PHY_HDR_SIGNAL PHY_HDR_SERVICE phy_hdr_length];
crc16_regs = ones(1,16);
for ii = 1:length(phy_hdr_protected)
    x = bitxor(crc16_regs(16), phy_hdr_protected(ii));

    crc16_regs = [x crc16_regs(1:4) ...
                  bitxor(x, crc16_regs(5)) ...
                  crc16_regs(6:11) ...
                  bitxor(x, crc16_regs(12)) ...
                  crc16_regs(13:15)];
end

phy_hdr_crc = fliplr(not(crc16_regs));

% Construct the PHY header bit sequence (192 bits total)
phy_hdr_bits = [PHY_HDR_SYNC PHY_HDR_SFD PHY_HDR_SIGNAL PHY_HDR_SERVICE phy_hdr_length phy_hdr_crc];

%%
% Calculate the MAC payload FCS
[mac_pyld_fcs, ~] = util_calc_80211_fcs(tx_sim.mac_payload(1:end-4));
tx_sim.mac_payload(end-3 : end) = mac_pyld_fcs;

mac_pyld_bits = fliplr(util_dec_to_bin_vec(tx_sim.mac_payload, 8));
mac_pyld_bits = reshape(mac_pyld_bits.', 1, []);

% Construct full packet bit sequence
pkt_bits = [phy_hdr_bits mac_pyld_bits];

% Scramble packet contents
pkt_bits_scr = NaN(size(pkt_bits));
scr_state = util_dec_to_bin_vec(SCR_INIT_STATE, 7);
for ii = 1:length(pkt_bits)
    x = bitxor(pkt_bits(ii), bitxor(scr_state(4), scr_state(7)));
    pkt_bits_scr(ii) = x;

    scr_state = [x, scr_state(1:6)];
end

if(payload_rate == 1)
    % Apply differential encoding to scrambled bits
    pkt_bits_scr_d = mod(cumsum(pkt_bits_scr), 2);

    % Modulate as BPSK (0 -> -1, 1 -> +1)
    pkt_syms = 2*pkt_bits_scr_d - 1;
else
    % Apply DBPSK modulation to PHY header bits
    hdr_bits_scr_d = mod(cumsum(pkt_bits_scr(1:192)), 2);
    hdr_syms = 2*hdr_bits_scr_d - 1;
    
    pyld_bits_scr = pkt_bits_scr(193:end);
    pyld_2b_vals = [2 1] * reshape(pyld_bits_scr, 2, []);
    
    modvec_dqpsk = [0 pi/2 3*pi/2 pi];
    dqpsk_mod = @(d) (modvec_dqpsk(d+1));

    % Compute phase change for each payload value
    pyld_phase_diffs = arrayfun(dqpsk_mod, pyld_2b_vals); 
    
    % Compute absolute phases for payload symbols, using final symbol
    %  of header as starting phase for the first payload phase diff
    pyld_phases = cumsum([angle(hdr_syms(end)) pyld_phase_diffs]);
    pyld_syms = exp(1i*pyld_phases(2:end));

    % Concatenate BSPK header and QPSK payload
    pkt_syms = [hdr_syms pyld_syms];
end
    
% Spread symbols by spreading sequence
spread_seq = [+1,  -1, +1, +1,  -1, +1, +1, +1,  -1,  -1,  -1]; 

pkt_syms_spread = fliplr(pkt_syms.' * fliplr(spread_seq));

pkt.syms = pkt_syms;

pkt.pkt_bits = pkt_bits;
pkt.pkt_bits_scr = pkt_bits_scr;

% Convert to 1-D waveform
pkt_wvfm11 = reshape(pkt_syms_spread.', 1, []);

% Pad with zeros to flush resampling filter
pkt_wvfm11 = [pkt_wvfm11 zeros(1, 50)];

if samp_rate == 20 
	pkt.wvfm = util_resamp20_11(pkt_wvfm11); 
else 
	pkt.wvfm = util_resamp40_11(pkt_wvfm11); 
end 

